/*
  httpdtalk.c  --  talk with a web server
  Copyright (C) 1999 Steffen Solyga <solyga@absinth.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include	"httpdtalk.h"

int
display_help( char* pn ) {
  fprintf( HELP_CHANNEL, "%s v%s (%s)\n", pn, VERSION_NUMBER, DATE_OF_LAST_MOD );
  fprintf( HELP_CHANNEL, "Talk with a web server.\n" );
  fprintf( HELP_CHANNEL, "Flowers & bug reports to %s.\n", MY_EMAIL_ADDRESS );
  fprintf( HELP_CHANNEL, "Usage: %s [options] [host]\n", pn );
  fprintf( HELP_CHANNEL, "  -h\t help, write this info to %s and exit sucessfully\n", HELP_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "  -H\t hexdump output\n" );
  fprintf( HELP_CHANNEL, "  -v\t raise verbosity level for %s\n", VERBOSE_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "  -V\t print version and compilation info to %s and exit sucessfully\n", VERSION_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "controllers:\n");
  fprintf( HELP_CHANNEL, "  -p port\t set port (default is %d)\n", DEFAULT_PORT );
  return( 0 );
}


int
display_version( char* pn ) {
  fprintf( VERSION_CHANNEL, "%s v%s (%s)\n", pn, VERSION_NUMBER, DATE_OF_LAST_MOD );
  fprintf( VERSION_CHANNEL, "compilation settings:\n" );
  fprintf( VERSION_CHANNEL, "  DEFAULT_HOST      :  %s\n", DEFAULT_HOST );
  fprintf( VERSION_CHANNEL, "  DEFAULT_PORT      :  %d\n", DEFAULT_PORT );
  fprintf( VERSION_CHANNEL, "  ANSWER_BUFFER_SIZE:  %d\n", ANSWER_BUFFER_SIZE );
  return( 0 );
}


ulong
get_ipaddr_n_by_name( char* hostname ) {
/* return binary address in network byte order */
  struct hostent* ent= gethostbyname( hostname );
  if( ent == NULL ) return( -1 );
  if( ent->h_addr_list[0] == NULL ) return( -1 );
  return( *( (ulong*)ent->h_addr_list[0] ) );
}


int
hexdump( FILE* channel, UCHAR* buf, int n ) {
#define	BPL	16		/* bytes per line */
  int nbw= 0;
  int i,j;			/* j - line, i - byte in buffer */
  for( j=0; j*BPL<n; j++ ) {
    nbw+= fprintf( channel, "%04x:", j*BPL );
    for( i=j*BPL; i<(j+1)*BPL; i++ ) {
      if( i%4 == 0 ) nbw+= fprintf( channel, " " );
      if( i<n ) nbw+= fprintf( channel, " %02x", buf[i] );
      else      nbw+= fprintf( channel, "   " );
    }
    nbw+= fprintf( channel, "  | " );
    for( i=j*BPL; i<MIN(n,(j+1)*BPL); i++ ) {
      nbw+= fprintf( channel, "%c", IS_PRINTABLE(buf[i])?buf[i]:'.' );
    }
    nbw+= fprintf( channel, "\n" );
  }
  return( nbw );
}


ssize_t
my_write( int fd, void *buf, size_t count ) {
/* write exactly count bytes to fd (else error) */
  ssize_t nbw;
  ssize_t tnbw= 0;
  size_t rem= count;
  unsigned char *p= (unsigned char*)buf;
  do {
    if( (nbw= write( fd, p+tnbw, rem )) == -1 ) return( -1 );
    tnbw+= nbw;
    rem-= nbw;
  } while( nbw>0 && rem>0 );
  return( tnbw );
}


int
my_recv( int s, void* buf, int len, unsigned int flags ) {
/* receive exactly len bytes from socket (else EOF or error) */
  UCHAR *p= buf;
  int nbr;
  int tnbr= 0;
  int rem= len;
  do {
    if( (nbr= recv( s, p+tnbr, rem, flags )) == -1 ) return( -1 );
    tnbr+= nbr;
    rem-= nbr;
  } while( nbr>0 && rem>0 );
  return( tnbr );
}


int
main( int argc, char** argv ) {
/*
SOLYGA --------------------
SOLYGA main() httpdtalk

SOLYGA started      : Wed Aug 25 10:06:20 CEST 1999 @beast
*/
  int retval= RETVAL_OK;
  int c;
  int verbose= 0;		/* switch: be verbose */
  int do_hexdump= 0;		/* switch: hexdump */
  char default_hostname[128]= DEFAULT_HOST;
  char* hostname= default_hostname;
  int port= DEFAULT_PORT;
  int sockfd= 0;
  int in_fd= STDIN_FILENO;
  int out_fd= STDOUT_FILENO;
  char default_out_filename[]= "(none)";
  char* out_fn= default_out_filename;
  struct sockaddr_in host;
  ulong ipaddr_n;
  UCHAR* qer= NULL;
  size_t qer_sze;
  UCHAR ans[ANSWER_BUFFER_SIZE+1];
  ssize_t nbr;			/* number of bytes read */
  ssize_t tnbr;			/* total number of bytes read */
  int nbts;			/* number of bytes to send */
  int nbs;			/* number of bytes sent */
  int nbc;			/* number of bytes received */
  int nbtc;			/* number of bytes to receive */
  int tnbc;			/* total number of bytes received */
  ssize_t nbw;			/* number of bytes written */
  int i;
  int proper_eoh;		/* flag: proper end-of-header (query) */
  int binary_answer= 0;		/* flag: non-ascii answer */
  int nlnl_read;		/* flag: 2 new-line read */
#if defined OS_TYPE_AIX
  extern int errno;
  extern char* sys_errlist[];
#endif
/***** process options *****/
  while( (c=getopt(argc,argv,"hHvVp:")) != EOF ) {
    switch( c ) {
      case 'p': /* set port */ {
        char* p;
        port= (int)strtol( optarg, &p, 0 );
        if( p == optarg ) {
          fprintf( ERROR_CHANNEL,
                   "%s: Invalid port number `%s'.\n",
                   *argv, optarg );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        break;
      }
      case 'h': /* display help and exit sucessfully */
        display_help(*argv);
        retval= RETVAL_OK; goto DIE_NOW;
      case 'H': /* hexdump output */
        do_hexdump= 1;
        break;
      case 'v': /* raise verbosity level */
        verbose++;
        break;
      case 'V': /* display version to VERSION_CHANNEL and exit sucessfully */
        display_version( *argv );
        retval= RETVAL_OK; goto DIE_NOW;
      case '?': /* refer to -h and exit unsucessfully */
        fprintf( ERROR_CHANNEL, "Try `%s -h' for more information.\n", *argv );
        retval= RETVAL_ERROR; goto DIE_NOW;
      default : /* program error */
        fprintf( ERROR_CHANNEL, "%s: Options bug! E-mail me at %s.\n",
                 *argv, MY_EMAIL_ADDRESS );
        retval= RETVAL_BUG; goto DIE_NOW;
    }
  }

/* prepare connexion */
  if( optind<argc ) hostname= argv[optind];
  if( (sockfd=socket(AF_INET,SOCK_STREAM,0)) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't create socket. %s.\n",
             *argv, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }
  if( verbose >= 1 ) {
    fprintf( VERBOSE_CHANNEL, "%s: Resolving hostname `%s'.\n",
             *argv, hostname );
  }
  if( (ipaddr_n=get_ipaddr_n_by_name( hostname )) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't resolve hostname `%s'. %s.\n",
             *argv, hostname, h_errlist[h_errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }
  if( verbose >= 2 ) {
    char str_ipaddr[]= "000.000.000.000";
    UCHAR* p= (UCHAR*)&ipaddr_n;
    sprintf( str_ipaddr, "%d.%d.%d.%d",
              p[0], p[1], p[2], p[3] );
    fprintf( VERBOSE_CHANNEL, "%s: IP address of `%s' is %s.\n",
             *argv, hostname, str_ipaddr );
  }

/* get query */
  if( verbose >= 1 ) {
    fprintf( VERBOSE_CHANNEL, "%s: Reading query from stdin.\n", *argv );
  }
  tnbr= 0;
  qer_sze= 0;
  nlnl_read= 0;
  while( 1 ) {
    if( qer_sze < tnbr+BUFFER_BASE_SIZE ) {
      qer_sze+= BUFFER_BASE_SIZE;
      if( (qer=realloc(qer,qer_sze)) == NULL ) {
        fprintf( ERROR_CHANNEL, "%s: Cannot allocate %d bytes for query.\n",
                 *argv, qer_sze );
        retval= RETVAL_MALLOC; goto DIE_NOW;
      }
    }
#ifdef	DEBUG
  fprintf( DEBUG_CHANNEL, "DEBUG: calling read(2)\n" );
#endif
    if( (nbr=read(in_fd,qer+tnbr,BUFFER_BASE_SIZE-1)) == -1 ) {
      fprintf( ERROR_CHANNEL, "%s: Read error on stdin. %s.\n",
               *argv, sys_errlist[errno] );
      retval= RETVAL_ERROR; goto DIE_NOW;
    }
    if( tnbr>0 && qer[tnbr] == '\n' ) nlnl_read= 1;
    tnbr+= nbr;
    if( nbr == 0  ||  nlnl_read ) break;
  }
  if( verbose >= 2 ) {
    fprintf( VERBOSE_CHANNEL, "%s: Query successfully read (%d bytes).\n",
            *argv, tnbr );
  }
#ifdef	DEBUG
  hexdump( STD_CHANNEL, qer, tnbr );
#endif

/* check query */
  nbts= tnbr;
  proper_eoh= 0;
  /* check for proper end-of-header, delete trailing garbage */
  for( i=0; i<tnbr-1; i++ ) {
    if( qer[i]=='\n' && qer[i+1]=='\n' ) {
      nbts= i+2;
      if( verbose >= 2  &&  nbts < tnbr) {
        fprintf( VERBOSE_CHANNEL, "%s: Deleting trailing garbage.\n", *argv );
      }
      proper_eoh= 1;
      break;
    }
  }
  if( !proper_eoh ) {
    /* read() left room for one additional byte */
    qer[tnbr] = '\n';
    proper_eoh= 1;
    nbts= tnbr+1;
    if( verbose >= 1 ) {
      fprintf( VERBOSE_CHANNEL, "%s: Added newline to query.\n", *argv );
    }
  }
  if( nbts <= 2 ) {
    fprintf( ERROR_CHANNEL, "%s: What a cool query! Exiting.\n", *argv );
    retval= RETVAL_QUERY; goto DIE_NOW;
  }


/* open connexion */
  host.sin_family= AF_INET;
  host.sin_port= htons(port);
  host.sin_addr.s_addr= ipaddr_n;
  if( verbose >= 1 ) {
    fprintf( VERBOSE_CHANNEL,
             "%s: Connecting to host `%s' at port %d.\n",
              *argv, hostname, port );
  }
  if( connect(sockfd,(struct sockaddr*)&host,sizeof(host)) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't connect to host `%s' port %d. %s.\n",
             *argv, hostname, port, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }

/* send query */
  if( verbose >= 1 ) {
    fprintf( VERBOSE_CHANNEL, "%s: Sending query (%d bytes).\n",
             *argv, nbts );
  }
  if( (nbs=send(sockfd,qer,nbts,0)) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't send query. %s.\n",
             *argv, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }
  if( verbose >= 2 ) {
    fprintf( VERBOSE_CHANNEL, "%s: Query successfully sent (%d bytes).\n",
             *argv, nbs );
  }

/* receive answer */
  if( verbose >= 1 ) {
    fprintf( VERBOSE_CHANNEL, "%s: Waiting for answer.\n",
             *argv );
  }
  nbtc= ANSWER_BUFFER_SIZE;
  tnbc= 0;
  while( 1 ) {
    if( (nbc=my_recv(sockfd,ans,nbtc,0)) == -1 ) {
      fprintf( ERROR_CHANNEL, "%s: Couldn't receive answer. %s.\n", 
               *argv, sys_errlist[errno] );
      retval= RETVAL_ERROR; goto DIE_NOW;
    }
    if( verbose >= 2 ) {
      fprintf( VERBOSE_CHANNEL, "%s: %d/%d bytes requested/received.\n",
               *argv, nbtc, nbc );
    }
    tnbc+= nbc;
    for( i=0; i<nbc; i++ ) if( NON_ASCII(ans[i]) ) binary_answer= 1;
    if( do_hexdump ) {
      ans[nbc]= '\0';
      hexdump( STD_CHANNEL, ans, nbc );
    }
    else {
      if( (nbw=my_write(out_fd,ans,nbc)) < nbc ) {
        fprintf( ERROR_CHANNEL, "%s: Couldn't write to ", *argv );
        if( out_fd == STDOUT_FILENO )
          fprintf( ERROR_CHANNEL, "stdout" );
        else
          fprintf( ERROR_CHANNEL, "`%s'", out_fn );
        fprintf( ERROR_CHANNEL, ". %s.\n", sys_errlist[errno] );
        retval= RETVAL_ERROR; goto DIE_NOW;
      }
    }
    if( nbc < nbtc ) break;
  }
  if( verbose >= 1 ) {
    fprintf( VERBOSE_CHANNEL,
             "%s: Answer successfully received (%d bytes, %s).\n",
             *argv, tnbc, binary_answer?"binary":"ascii" );
  }

/* close connexion */
  if( close( sockfd ) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Couldn't close socket. %s.\n",
             *argv, sys_errlist[errno] );
    retval= RETVAL_ERROR; goto DIE_NOW;
  }

DIE_NOW:
  free( qer );
  close( sockfd );
  exit( retval );
}
